@@ -26,7 +26,7 @@ void chmac PARAMS((void));
 
 /* Text to macro / Macro to text */
 MACRO *mparse PARAMS((MACRO *m, unsigned char *buf, int *sta));
-char *mtext PARAMS((char *s, MACRO *m));
+unsigned char *mtext PARAMS((unsigned char *s, MACRO *m));
 
 /* Execute a macro */
 extern MACRO *curmacro;
