@@ -21,11 +21,11 @@ void *alitem(void *list, int itemsize)
 
 	if (qempty(STDITEM, link, freelist)) {
 		STDITEM *i = (STDITEM *) joe_malloc(itemsize * 16);
-		STDITEM *z = (STDITEM *) ((char *) i + itemsize * 16);
+		STDITEM *z = (STDITEM *) ((unsigned char *) i + itemsize * 16);
 
 		while (i != z) {
 			enquef(STDITEM, link, freelist, i);
-			i = (STDITEM *) ((char *) i + itemsize);
+			i = (STDITEM *) ((unsigned char *) i + itemsize);
 		}
 	}
 	return (void *) deque_f(STDITEM, link, freelist->link.prev);
