@@ -10,6 +10,8 @@
 
 #include "utils.h"
 #include "vs.h"
+#include "charmap.h"
+#include "utf8.h"
 #include "w.h"
 
 static void dispqw(QW *qw)
@@ -28,7 +30,17 @@ static void dispqw(QW *qw)
 
 	/* Generate prompt */
 	w->t->t->updtab[w->y] = 1;
-	gentxt(w->t->t, w->x, w->y, qw->promptofst, qw->prompt, qw->promptlen, 1);
+	genfield(w->t->t,
+	         w->t->t->scrn + w->y * w->t->t->co + w->x,
+	         w->t->t->attr + w->y * w->t->t->co + w->x,
+	         w->x,
+	         w->y,
+	         qw->promptofst,
+	         qw->prompt,
+	         qw->promptlen,
+	         0,
+	         w->w - w->x,
+	         1);
 }
 
 static void dispqwn(QW *qw)
@@ -51,11 +63,23 @@ static void dispqwn(QW *qw)
 
 	/* Generate prompt */
 	w->t->t->updtab[w->y] = 1;
-	gentxt(w->t->t, w->x, w->y, qw->promptofst, qw->prompt, qw->promptlen, 1);
+	genfield(w->t->t,
+	         w->t->t->scrn + w->y * w->t->t->co + w->x,
+	         w->t->t->attr + w->y * w->t->t->co + w->x,
+	         w->x,
+	         w->y,
+	         qw->promptofst,
+	         qw->prompt,
+	         qw->promptlen,
+	         0,
+	         w->w - w->x,
+	         1);
 }
 
 /* When user hits a key in a query window */
 
+struct utf8_sm qw_sm;
+
 static int utypeqw(QW *qw, int c)
 {
 	W *win;
@@ -64,6 +88,12 @@ static int utypeqw(QW *qw, int c)
 	int (*func) ();
 	void *object = qw->object;
 
+	if (locale_map->type) {
+		c = utf8_decode(&qw_sm, c);
+		if (c<0)
+			return 0;
+	}
+
 	win = qw->parent->win;
 	func = qw->func;
 	vsrm(qw->prompt);
@@ -91,7 +121,7 @@ static int abortqw(QW *qw)
 }
 
 static WATOM watomqw = {
-	"query",
+	US "query",
 	dispqw,
 	NULL,
 	abortqw,
@@ -105,7 +135,7 @@ static WATOM watomqw = {
 };
 
 static WATOM watqwn = {
-	"querya",
+	US "querya",
 	dispqwn,
 	NULL,
 	abortqw,
@@ -119,7 +149,7 @@ static WATOM watqwn = {
 };
 
 static WATOM watqwsr = {
-	"querysr",
+	US "querysr",
 	dispqwn,
 	NULL,
 	abortqw,
@@ -134,7 +164,7 @@ static WATOM watqwsr = {
 
 /* Create a query window */
 
-QW *mkqw(W *w, char *prompt, int len, int (*func) (/* ??? */), int (*abrt) (/* ??? */), void *object, int *notify)
+QW *mkqw(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (*abrt) (/* ??? */), void *object, int *notify)
 {
 	W *new;
 	QW *qw;
@@ -161,7 +191,7 @@ QW *mkqw(W *w, char *prompt, int len, int (*func) (/* ??? */), int (*abrt) (/* ?
 /* Same as above, but cursor is left in original window */
 /* For Ctrl-Meta thing */
 
-QW *mkqwna(W *w, char *prompt, int len, int (*func) (/* ??? */), int (*abrt) (/* ??? */), void *object, int *notify)
+QW *mkqwna(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (*abrt) (/* ??? */), void *object, int *notify)
 {
 	W *new;
 	QW *qw;
@@ -188,7 +218,7 @@ QW *mkqwna(W *w, char *prompt, int len, int (*func) (/* ??? */), int (*abrt) (/*
 /* Same as above, but cursor is left in original window */
 /* For search and replace thing */
 
-QW *mkqwnsr(W *w, char *prompt, int len, int (*func) (/* ??? */), int (*abrt) (/* ??? */), void *object, int *notify)
+QW *mkqwnsr(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (*abrt) (/* ??? */), void *object, int *notify)
 {
 	W *new;
 	QW *qw;
