@@ -27,6 +27,13 @@
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
 #endif
+
+#ifdef HAVE_OPENPTY
+#ifdef HAVE_PTY_H
+#include <pty.h>
+#endif
+#endif
+
 #include <errno.h>
 #include <string.h>
 #include <stdlib.h>
@@ -137,7 +144,7 @@ static struct ltchars oltarg;
 
 /* Output buffer, index and size */
 
-char *obuf = NULL;
+unsigned char *obuf = NULL;
 int obufp = 0;
 int obufsiz;
 
@@ -189,14 +196,13 @@ static int mpxfd;		/* Editor reads packets from this fd */
 static int mpxsfd;		/* Clients send packets to this fd */
 
 static int nmpx = 0;
-static int accept = MAXINT;	/* =MAXINT if we have last packet */
-				/* FIXME: overloaded meaning of MAXINT */
+static int accept = NO_MORE_DATA;	/* =-1 if we have last packet */
 
 struct packet {
 	MPX *who;
 	int size;
 	int ch;
-	char data[1024];
+	unsigned char data[1024];
 } pack;
 
 MPX asyncs[NPROC];
@@ -258,7 +264,6 @@ extern int dostaupd;
 static RETSIGTYPE dotick(int unused)
 {
 	ticked = 1;
-	dostaupd = 1;
 }
 
 void tickoff(void)
@@ -299,7 +304,6 @@ void ttopnn(void)
 #ifdef SIGWINCH
 			joe_set_signal(SIGWINCH, winchd);
 #endif
-			tickon();
 		}
 	}
 
@@ -385,7 +389,7 @@ void ttopnn(void)
 	}
 	if (!obufsiz)
 		obufsiz = 1;
-	obuf = (char *) joe_malloc(obufsiz);
+	obuf = (unsigned char *) joe_malloc(obufsiz);
 }
 
 /* Close terminal */
@@ -498,7 +502,6 @@ int ttflsh(void)
 			while (!yep)
 				pauseit();
 			unmaskit();
-			tickon();
 		} else
 			joe_write(fileno(termout), obuf, obufp);
 
@@ -517,14 +520,14 @@ int ttflsh(void)
 	}
 
 	/* Ack previous packet */
-	if (ackkbd != -1 && accept != MAXINT && !have) {
-		char c = 0;
+	if (ackkbd != -1 && accept != NO_MORE_DATA && !have) {
+		unsigned char c = 0;
 
 		if (pack.who && pack.who->func)
 			joe_write(pack.who->ackfd, &c, 1);
 		else
 			joe_write(ackkbd, &c, 1);
-		accept = MAXINT;
+		accept = NO_MORE_DATA;
 	}
 
 	/* Check for typeahead or next packet */
@@ -558,11 +561,23 @@ int ttflsh(void)
 
 void mpxdied(MPX *m);
 
+long last_time;
+
 int ttgetc(void)
 {
 	int stat;
+	long new_time;
+
+
+	tickon();
 
       loop:
+	new_time = time(NULL);
+	if (new_time != last_time) {
+		last_time = new_time;
+		dostaupd = 1;
+		ticked = 1;
+	}
 	ttflsh();
 	while (winched) {
 		winched = 0;
@@ -590,7 +605,7 @@ int ttgetc(void)
 		}
 		have = 0;
 		if (pack.who) {	/* Got bknd input */
-			if (accept != MAXINT) {
+			if (accept != NO_MORE_DATA) {
 				if (pack.who->func) {
 					pack.who->func(pack.who->object, pack.data, pack.size);
 					edupd(1);
@@ -599,9 +614,12 @@ int ttgetc(void)
 				mpxdied(pack.who);
 			goto loop;
 		} else {
-			if (accept != MAXINT)
+			if (accept != NO_MORE_DATA) {
+				tickoff();
 				return accept;
+			}
 			else {
+				tickoff();
 				ttsig(0);
 				return 0;
 			}
@@ -617,12 +635,13 @@ int ttgetc(void)
 				ttsig(0);
 		}
 	}
+	tickoff();
 	return havec;
 }
 
 /* Write string to output */
 
-void ttputs(char *s)
+void ttputs(unsigned char *s)
 {
 	while (*s) {
 		obuf[obufp++] = *s++;
@@ -661,13 +680,15 @@ void ttgtsz(int *x, int *y)
 #endif
 }
 
-void ttshell(char *cmd)
+void ttshell(unsigned char *cmd)
 {
 	int x, omode = ttymode;
-	char *s = getenv("SHELL");
+	unsigned char *s = (unsigned char *)getenv("SHELL");
 
-	if (!s)
-		return;
+	if (!s) {
+		s = US "/bin/sh";
+		/* return; */
+	}
 	ttclsn();
 	if ((x = fork()) != 0) {
 		if (x != -1)
@@ -677,44 +698,22 @@ void ttshell(char *cmd)
 	} else {
 		signrm();
 		if (cmd)
-			execl(s, s, "-c", cmd, NULL);
+			execl((char *)s, (char *)s, "-c", cmd, NULL);
 		else {
 			fprintf(stderr, "You are at the command shell.  Type 'exit' to return\n");
-			execl(s, s, NULL);
+			execl((char *)s, (char *)s, NULL);
 		}
 		_exit(0);
 	}
 }
 
-void ttsusp(void)
-{
-	int omode;
+/* Create keyboard task */
 
-	tickoff();
-#ifdef SIGTSTP
-	omode = ttymode;
-	ttclsn();
-	fprintf(stderr, "You have suspended the program.  Type 'fg' to return\n");
-	kill(0, SIGTSTP);
-	if (ackkbd != -1)
-		kill(kbdpid, SIGCONT);
-	if (omode)
-		ttopnn();
-#else
-	ttshell(NULL);
-#endif
-	tickon();
-}
-
-static void mpxstart(void)
+static void mpxresume(void)
 {
 	int fds[2];
-
 	pipe(fds);
-	mpxfd = fds[0];
-	mpxsfd = fds[1];
-	pipe(fds);
-	accept = MAXINT;
+	accept = NO_MORE_DATA;
 	have = 0;
 	if (!(kbdpid = fork())) {
 		close(fds[1]);
@@ -725,7 +724,7 @@ static void mpxstart(void)
 			pack.who = 0;
 			sta = joe_read(fileno(termin), &c, 1);
 			if (sta == 0)
-				pack.ch = MAXINT;
+				pack.ch = NO_MORE_DATA;
 			else
 				pack.ch = c;
 			pack.size = 0;
@@ -737,12 +736,69 @@ static void mpxstart(void)
 	ackkbd = fds[1];
 }
 
+/* Kill keyboard task */
+
+static void mpxsusp(void)
+{
+	if (ackkbd!=-1) {
+		kill(kbdpid, 9);
+		while (wait(NULL) < 0 && errno == EINTR)
+			/* do nothing */;
+		close(ackkbd);
+	}
+}
+
+/* We used to leave the keyboard copy task around during suspend, but
+   Cygwin gets confused when two processes are waiting for input and you
+   change the tty from raw to cooked (on the call to ttopnn()): the keyboard
+   process was stuck in cooked until he got a carriage return- then he
+   switched back to raw (he's supposed to switch to raw without waiting for
+   the end of line). Probably this should be done for ttshell() as well. */
+
+void ttsusp(void)
+{
+	int omode;
+
+#ifdef SIGTSTP
+	omode = ttymode;
+	mpxsusp();
+	ttclsn();
+	fprintf(stderr, "You have suspended the program.  Type 'fg' to return\n");
+	kill(0, SIGTSTP);
+#ifdef junk
+	/* Hmmm... this should not have been necessary */
+	if (ackkbd != -1)
+		kill(kbdpid, SIGCONT);
+#endif
+	if (omode)
+		ttopnn();
+	if (ackkbd!= -1)
+		mpxresume();
+#else
+	ttshell(NULL);
+#endif
+}
+
+/* Stuff for asynchronous I/O multiplexing.  We do not use streams or
+   select() because joe needs to work on versions of UNIX which predate
+   these calls.  Instead, when there is multiple async sources, we use
+   helper processes which packetize data from the sources.  A header on each
+   packet indicates the source.  There is no guarentee that packets getting
+   written to the same pipe don't get interleaved, but you can reasonable
+   rely on it with small packets. */
+
+static void mpxstart(void)
+{
+	int fds[2];
+	pipe(fds);
+	mpxfd = fds[0];
+	mpxsfd = fds[1];
+	mpxresume();
+}
+
 static void mpxend(void)
 {
-	kill(kbdpid, 9);
-	while (wait(NULL) < 0 && errno == EINTR)
-		/* do nothing */;
-	close(ackkbd);
+	mpxsusp();
 	ackkbd = -1;
 	close(mpxfd);
 	close(mpxsfd);
@@ -754,37 +810,61 @@ static void mpxend(void)
  * string in static buffer or NULL if couldn't get a pair.
  */
 
+#ifdef __svr4__
+#define USEPTMX 1
+#else
+#ifdef __CYGWIN__
+#define USEPTMX 1
+#endif
+#endif
+
 #ifdef sgi
 
 /* Newer sgi machines can do it the __svr4__ way, but old ones can't */
 
 extern char *_getpty();
 
-static char *getpty(int *ptyfd)
+static unsigned char *getpty(int *ptyfd)
 {
-	return _getpty(ptyfd, O_RDWR, 0600, 0);
+	return (unsigned char *)_getpty(ptyfd, O_RDWR, 0600, 0);
 }
 
 #else
-#ifdef __svr4__
+#ifdef USEPTMX
 
 /* Strange streams way */
 
 extern char *ptsname();
 
-static char *getpty(int *ptyfd)
+static unsigned char *getpty(int *ptyfd)
 {
 	int fdm;
-	char *name;
+	unsigned char *name;
 
 	*ptyfd = fdm = open("/dev/ptmx", O_RDWR);
 	grantpt(fdm);
 	unlockpt(fdm);
-	return ptsname(fdm);
+	return (unsigned char *)ptsname(fdm);
 }
 
 #else
+#ifdef HAVE_OPENPTY
 
+/* BSD function, present in libc5 and glibc2 */
+
+static unsigned char *getpty(int *ptyfd)
+{
+	int fdm;
+	static unsigned char name[32];
+	int ttyfd;
+
+        if (openpty(ptyfd, &ttyfd, name, NULL, NULL) == 0)
+           return(name);
+        else
+	   return (NULL);
+}
+
+#else
 /* The normal way: for each possible pty/tty pair, try to open the pty and
  * then the corresponding tty.  If both could be opened, close them both and
  * then re-open the pty.  If that succeeded, return with the opened pty and the
@@ -797,46 +877,46 @@ static char *getpty(int *ptyfd)
  * process and the process gets to be the session leader.
  */
 
-static char *getpty(int *ptyfd)
+static unsigned char *getpty(int *ptyfd)
 {
 	int x, fd;
-	char *orgpwd = pwd();
-	static char **ptys = NULL;
-	static char *ttydir;
-	static char *ptydir;
-	static char ttyname[32];
+	unsigned char *orgpwd = pwd();
+	static unsigned char **ptys = NULL;
+	static unsigned char *ttydir;
+	static unsigned char *ptydir;
+	static unsigned char ttyname[32];
 
 	if (!ptys) {
-		ttydir = "/dev/pty/";
-		ptydir = "/dev/ptym/";	/* HPUX systems */
-		if (chpwd(ptydir) || !(ptys = rexpnd("pty*")))
+		ttydir = US "/dev/pty/";
+		ptydir = US "/dev/ptym/";	/* HPUX systems */
+		if (chpwd(ptydir) || !(ptys = rexpnd(US "pty*")))
 			if (!ptys) {
-				ttydir = ptydir = "/dev/";	/* Everyone else */
+				ttydir = ptydir = US "/dev/";	/* Everyone else */
 				if (!chpwd(ptydir))
-					ptys = rexpnd("pty*");
+					ptys = rexpnd(US "pty*");
 			}
 	}
 	chpwd(orgpwd);
 
 	if (ptys)
 		for (fd = 0; ptys[fd]; ++fd) {
-			strcpy(ttyname, ptydir);
-			strcat(ttyname, ptys[fd]);
-			if ((*ptyfd = open(ttyname, O_RDWR)) >= 0) {
+			strcpy((char *)ttyname, (char *)ptydir);
+			strcat((char *)ttyname, (char  *)(ptys[fd]));
+			if ((*ptyfd = open((char *)ttyname, O_RDWR)) >= 0) {
 				ptys[fd][0] = 't';
-				strcpy(ttyname, ttydir);
-				strcat(ttyname, ptys[fd]);
+				strcpy((char *)ttyname, (char *)ttydir);
+				strcat((char *)ttyname, (char *)(ptys[fd]));
 				ptys[fd][0] = 'p';
-				x = open(ttyname, O_RDWR);
+				x = open((char *)ttyname, O_RDWR);
 				if (x >= 0) {
 					close(x);
 					close(*ptyfd);
-					strcpy(ttyname, ptydir);
-					strcat(ttyname, ptys[fd]);
-					*ptyfd = open(ttyname, O_RDWR);
+					strcpy((char *)ttyname, (char *)ptydir);
+					strcat((char *)ttyname, (char *)(ptys[fd]));
+					*ptyfd = open((char *)ttyname, O_RDWR);
 					ptys[fd][0] = 't';
-					strcpy(ttyname, ttydir);
-					strcat(ttyname, ptys[fd]);
+					strcpy((char *)ttyname, (char *)ttydir);
+					strcat((char *)ttyname, (char *)(ptys[fd]));
 					ptys[fd][0] = 'p';
 					return ttyname;
 				} else
@@ -846,13 +926,18 @@ static char *getpty(int *ptyfd)
 	return NULL;
 }
 
+#endif
 #endif
 #endif
 
+/* Shell dies signal handler.  Puts pty in non-block mode so
+ * that read returns with <1 when all data from process has
+ * been read. */
 int dead = 0;
-
+int death_fd;
 static RETSIGTYPE death(int unused)
 {
+	fcntl(death_fd,F_SETFL,O_NDELAY);
 	wait(NULL);
 	dead = 1;
 }
@@ -861,17 +946,17 @@ static RETSIGTYPE death(int unused)
 #define SIGCHLD SIGCLD
 #endif
 
-/* Build a new environment */
+/* Build a new environment, but replace one variable */
 
-extern char **mainenv;
+extern unsigned char **mainenv;
 
-static char **newenv(char **old, char *s)
+static unsigned char **newenv(unsigned char **old, unsigned char *s)
 {
-	char **new;
+	unsigned char **new;
 	int x, y, z;
 
 	for (x = 0; old[x]; ++x) ;
-	new = (char **) joe_malloc((x + 2) * sizeof(char *));
+	new = (unsigned char **) joe_malloc((x + 2) * sizeof(unsigned char *));
 
 	for (x = 0, y = 0; old[x]; ++x) {
 		for (z = 0; s[z] != '='; ++z)
@@ -889,71 +974,125 @@ static char **newenv(char **old, char *s)
 	return new;
 }
 
-MPX *mpxmk(int *ptyfd, char *cmd, char **args, void (*func) (/* ??? */), void *object, void (*die) (/* ??? */), void *dieobj)
+/* Create a shell process */
+
+MPX *mpxmk(int *ptyfd, unsigned char *cmd, unsigned char **args, void (*func) (/* ??? */), void *object, void (*die) (/* ??? */), void *dieobj)
 {
+	unsigned char buf[80];
 	int fds[2];
 	int comm[2];
 	pid_t pid;
 	int x;
 	MPX *m;
-	char *name;
+	unsigned char *name;
 
+	/* Get pty/tty pair */
 	if (!(name = getpty(ptyfd)))
 		return NULL;
+
+	/* Find free slot */
 	for (x = 0; x != NPROC; ++x)
 		if (!asyncs[x].func) {
 			m = asyncs + x;
-			goto ok;
+			break;
 		}
-	return NULL;
-      ok:
+	if (x==NPROC)
+		return NULL;
+
+	/* Fixes cygwin console bug: if you fork() with inverse video he assumes you want
+	 * ESC [ 0 m to keep it in inverse video from then on. */
+	set_attr(maint->t,0);
+
+	/* Flush output */
 	ttflsh();
+
+	/* Bump no. current async inputs to joe */
 	++nmpx;
+
+	/* Start input multiplexer */
 	if (ackkbd == -1)
 		mpxstart();
+
+	/* Remember callback function */
 	m->func = func;
 	m->object = object;
 	m->die = die;
 	m->dieobj = dieobj;
+
+	/* Acknowledgement pipe */
 	pipe(fds);
-	pipe(comm);
 	m->ackfd = fds[1];
+
+	/* PID number pipe */
+	pipe(comm);
+
+	/* Create processes... */
 	if (!(m->kpid = fork())) {
+		/* This process copies data from shell to joe */
+		/* After each packet it sends to joe it waits for
+		   an acknowledgement from joe so that it can not get
+		   too far ahead with buffering */
+
+		/* Close joe side of pipes */
 		close(fds[1]);
 		close(comm[0]);
+
+		/* Flag which indicates child died */
 		dead = 0;
+		death_fd = *ptyfd;
 		joe_set_signal(SIGCHLD, death);
 
 		if (!(pid = fork())) {
+			/* This process becomes the shell */
 			signrm();
+
+			/* Close pty (we only need tty) */
 			close(*ptyfd);
 
+			/* All of this stuff is for disassociating ourself from
+			   controlling tty (session leader) and starting a new
+			   session.  This is the most non-portable part of UNIX- second
+			   only to pty/tty pair creation. */
+#ifndef HAVE_LOGIN_TTY
+
 #ifdef TIOCNOTTY
 			x = open("/dev/tty", O_RDWR);
 			ioctl(x, TIOCNOTTY, 0);
 #endif
 
+			setsid();	/* I think you do setprgp(0,0) on systems with no setsid() */
 #ifndef SETPGRP_VOID
 			setpgrp(0, 0);
 #else
 			setpgrp();
 #endif
 
+#endif
+			/* Close all fds */
 			for (x = 0; x != 32; ++x)
 				close(x);	/* Yes, this is quite a kludge... all in the
 						   name of portability */
 
-			if ((x = open(name, O_RDWR)) != -1) {	/* Standard input */
-				char **env = newenv(mainenv, "TERM=");
+			/* Open the TTY */
+			if ((x = open((char *)name, O_RDWR)) != -1) {	/* Standard input */
+				unsigned char **env = newenv(mainenv, US "TERM=");
 
+#ifdef HAVE_LOGIN_TTY
+				login_tty(x);
+
+#else
+				/* This tells the fd that it's a tty (I think) */
 #ifdef __svr4__
 				ioctl(x, I_PUSH, "ptem");
 				ioctl(x, I_PUSH, "ldterm");
 #endif
+
+				/* Open stdout, stderr */
 				dup(x);
 				dup(x);	/* Standard output, standard error */
 				/* (yes, stdin, stdout, and stderr must all be open for reading and
 				 * writing.  On some systems the shell assumes this */
+#endif
 
 				/* We could probably have a special TTY set-up for JOE, but for now
 				 * we'll just use the TTY setup for the TTY was was run on */
@@ -970,27 +1109,52 @@ MPX *mpxmk(int *ptyfd, char *cmd, char **args, void (*func) (/* ??? */), void *o
 #endif
 
 				/* Execute the shell */
-				execve(cmd, args, env);
+				execve((char *)cmd, (char **)args, (char **)env);
+
+				/* If shell didn't execute */
+				joe_snprintf_1((char *)buf,sizeof(buf),"Couldn't execute shell '%s'\n",cmd);
+				write(0,(char *)buf,strlen((char *)buf));
+				sleep(1);
 			}
 
 			_exit(0);
 		}
+
+		/* Tell JOE PID of shell */
 		joe_write(comm[1], &pid, sizeof(pid));
 
+		/* sigpipe should be ignored here. */
+
+		/* This process copies data from shell to JOE until EOF.  It creates a packet
+		   for each data */
+
+
+		/* We don't really get EOF from a pty- it would just wait forever
+		   until someone else writes to the tty.  So: when the shell
+		   dies, the child died signal handler death() puts pty in non-block
+		   mode.  This allows us to read any remaining data- then
+		   read returns 0 and we know we're done. */
+
 	      loop:
 		pack.who = m;
 		pack.ch = 0;
-		if (dead)
-			pack.size = 0;
-		else
-			pack.size = read(*ptyfd, pack.data, 1024);
+
+		/* Read data from process */
+		pack.size = joe_read(*ptyfd, pack.data, 1024);
+
+		/* On SUNOS 5.8, the very first read from the pty returns 0 for some reason */
+		if (!pack.size)
+			pack.size = joe_read(*ptyfd, pack.data, 1024);
+
 		if (pack.size > 0) {
+			/* Send data to JOE, wait for ack */
 			joe_write(mpxsfd, &pack, sizeof(struct packet) - 1024 + pack.size);
 
 			joe_read(fds[0], &pack, 1);
 			goto loop;
 		} else {
-			pack.ch = MAXINT;
+			/* Shell died: return */
+			pack.ch = NO_MORE_DATA;
 			pack.size = 0;
 			joe_write(mpxsfd, &pack, sizeof(struct packet) - 1024);
 
@@ -999,8 +1163,11 @@ MPX *mpxmk(int *ptyfd, char *cmd, char **args, void (*func) (/* ??? */), void *o
 	}
 	joe_read(comm[0], &m->pid, sizeof(m->pid));
 
+	/* We only need comm once */
 	close(comm[0]);
 	close(comm[1]);
+
+	/* Close other side of copy process pipe */
 	close(fds[0]);
 	return m;
 }
