@@ -9,6 +9,9 @@
 #include "types.h"
 
 #include <unistd.h>
+#ifdef HAVE_STDLIB_H
+#include <stdlib.h>
+#endif
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
@@ -25,6 +28,8 @@
 #include "uedit.h"
 #include "utils.h"
 #include "vs.h"
+#include "utf8.h"
+#include "charmap.h"
 #include "w.h"
 
 /* Global options */
@@ -89,12 +94,21 @@ int upop(BW *bw)
 }
 
 /* Return true if markb/markk are valid */
+/* If r is set, swap markb with markk if necessary */
+
+int autoswap;
 
 int markv(int r)
 {
-	if (markb && markk && markb->b == markk->b && markk->byte > markb->byte && (!square || markk->xcol > markb->xcol))
+	if (markb && markk && markb->b == markk->b && markk->byte > markb->byte && (!square || markk->xcol > markb->xcol)) {
 		return 1;
-	else
+	} else if(autoswap && r && markb && markk && markb->b == markk->b && markb->byte > markk->byte && (!square || markk->xcol < markb->xcol)) {
+		P *p = pdup(markb);
+		prm(markb); markb=0; pdupown(markk, &markb);
+		prm(markk); markk=0; pdupown(p, &markk);
+		prm(p);
+		return 1;
+	} else
 		return 0;
 }
 
@@ -190,7 +204,7 @@ int ptabrect(P *org, long int height, long int right)
 		while ((c = pgetc(p)) != NO_MORE_DATA && c != '\n') {
 			if (c == '\t') {
 				prm(p);
-				return 1;
+				return '\t';
 			} else if (piscol(p) > right)
 				break;
 		}
@@ -198,7 +212,7 @@ int ptabrect(P *org, long int height, long int right)
 			pnextl(p);
 	}
 	prm(p);
-	return 0;
+	return ' ';
 }
 
 /* Insert rectangle */
@@ -254,6 +268,76 @@ int udrop(BW *bw)
 	return 0;
 }
 
+int ubegin_marking(BW *bw)
+{
+	if (marking)
+		/* We're marking now... don't stop */
+		return 0;
+	else if (markv(0) && bw->cursor->b==markb->b)
+		/* Try to extend current block */
+		if (bw->cursor->byte==markb->byte) {
+			pset(markb,markk);
+			prm(markk); markk=0;
+			marking = 1;
+			return 0;
+		} else if(bw->cursor->byte==markk->byte) {
+			prm(markk); markk=0;
+			marking = 1;
+			return 0;
+		}
+	/* Start marking - no message */
+	prm(markb); markb=0;
+	prm(markk); markk=0;
+	updall();
+	marking = 1;
+	return umarkb(bw);
+}
+
+int utoggle_marking(BW *bw)
+{
+	if (markv(0) && bw->cursor->b==markb->b && bw->cursor->byte>=markb->byte && bw->cursor->byte<=markk->byte) {
+		/* Just clear selection */
+		prm(markb); markb=0;
+		prm(markk); markk=0;
+		updall();
+		marking = 0;
+		msgnw(bw->parent, US "Selection cleared.");
+		return 0;
+	} else if (markk) {
+		/* Clear selection and start new one */
+		prm(markb); markb=0;
+		prm(markk); markk=0;
+		updall();
+		marking = 1;
+		msgnw(bw->parent, US "Selection started.");
+		return umarkb(bw);
+	} else if (markb && markb->b==bw->cursor->b) {
+		marking = 0;
+		if (bw->cursor->byte<markb->byte) {
+			pdupown(markb, &markk);
+			prm(markb); markb=0;
+			pdupown(bw->cursor, &markb);
+			markb->xcol = bw->cursor->xcol;
+		} else {
+			pdupown(bw->cursor, &markk);
+			markk->xcol = bw->cursor->xcol;
+		}
+		updall(); /* Because other windows could be changed */
+		return 0;
+	} else {
+		marking = 1;
+		msgnw(bw->parent, US "Selection started.");
+		return umarkb(bw);
+	}
+}
+
+int uselect(BW *bw)
+{
+	if (!markb)
+		umarkb(bw);
+	return 0;
+}
+
 /* Set end */
 
 int umarkk(BW *bw)
@@ -351,7 +435,7 @@ int ublkdel(BW *bw)
 		if (lightoff)
 			unmark(bw);
 	} else {
-		msgnw(bw->parent, "No block");
+		msgnw(bw->parent, US "No block");
 		return -1;
 	}
 	return 0;
@@ -388,7 +472,7 @@ int ublkmove(BW *bw)
 {
 	if (markv(1)) {
 		if (markb->b->rdonly) {
-			msgnw(bw->parent, "Read only");
+			msgnw(bw->parent, US "Read only");
 			return -1;
 		}
 		if (square) {
@@ -406,7 +490,7 @@ int ublkmove(BW *bw)
 				/* If cursor was in block, blkdel moves it to left edge of block, so fix it
 				 * back to its original place here */
 				pcol(bw->cursor, ocol);
-				pfill(bw->cursor, ocol, 0);
+				pfill(bw->cursor, ocol, ' ');
 				pdelrect(bw->cursor, height, piscol(bw->cursor) + width);
 			} else if (update_xcol)
 				/* If cursor was to right of block, xcol was not properly updated */
@@ -439,7 +523,7 @@ int ublkmove(BW *bw)
 			return 0;
 		}
 	}
-	msgnw(bw->parent, "No block");
+	msgnw(bw->parent, US "No block");
 	return -1;
 }
 
@@ -483,7 +567,7 @@ int ublkcpy(BW *bw)
 			return 0;
 		}
 	} else {
-		msgnw(bw->parent, "No block");
+		msgnw(bw->parent, US "No block");
 		return -1;
 	}
 }
@@ -491,7 +575,7 @@ int ublkcpy(BW *bw)
 /* Write highlighted block to a file */
 /* This is called by ublksave in ufile.c */
 
-int dowrite(BW *bw, char *s, void *object, int *notify)
+int dowrite(BW *bw, unsigned char *s, void *object, int *notify)
 {
 	if (notify)
 		*notify = 1;
@@ -503,8 +587,8 @@ int dowrite(BW *bw, char *s, void *object, int *notify)
 					  markk->line - markb->line + 1,
 					  markk->xcol);
 
-			if ((fl = bsave(tmp->bof, s, tmp->eof->byte)) != 0) {
-				msgnw(bw->parent, msgs[5 + fl]);
+			if ((fl = bsave(tmp->bof, s, tmp->eof->byte, 0)) != 0) {
+				msgnw(bw->parent, msgs[-fl]);
 				ret = -1;
 			}
 			brm(tmp);
@@ -516,8 +600,8 @@ int dowrite(BW *bw, char *s, void *object, int *notify)
 			int fl;
 			int ret = 0;
 
-			if ((fl = bsave(markb, s, markk->byte - markb->byte)) != 0) {
-				msgnw(bw->parent, msgs[5 + fl]);
+			if ((fl = bsave(markb, s, markk->byte - markb->byte, 0)) != 0) {
+				msgnw(bw->parent, msgs[-fl]);
 				ret = -1;
 			}
 			if (lightoff)
@@ -527,7 +611,7 @@ int dowrite(BW *bw, char *s, void *object, int *notify)
 		}
 	} else {
 		vsrm(s);
-		msgnw(bw->parent, "No block");
+		msgnw(bw->parent, US "No block");
 		return -1;
 	}
 }
@@ -551,8 +635,9 @@ void setindent(BW *bw)
 			goto done;
 		else
 			p_goto_bol(p);
-	} while (pisindent(p) >= indent && !pisblank(p));
+	} while (pisindent(p) >= indent || pisblank(p));
 	pnextl(p);
+	/* Maybe skip blank lines at beginning */
       done:
 	p_goto_bol(p);
 	p->xcol = piscol(p);
@@ -564,8 +649,8 @@ void setindent(BW *bw)
 	do {
 		if (!pnextl(q))
 			break;
-	} while (pisindent(q) >= indent && !pisblank(q));
-
+	} while (pisindent(q) >= indent || pisblank(q));
+	/* Maybe skip blank lines at end */
 	if (markk)
 		prm(markk);
 	q->xcol = piscol(q);
@@ -575,6 +660,55 @@ void setindent(BW *bw)
 	updall();
 }
 
+/* Purity check */
+/* Verifies that at least n indentation characters (for non-blank lines) match c */
+/* If n is 0 (for urindent), this fails if c is space but indentation begins with tab */
+
+int purity_check(int c, int n)
+{
+	P *p = pdup(markb);
+	while (p->byte < markk->byte) {
+		int x;
+		p_goto_bol(p);
+		if (!n && c==' ' && brc(p)=='\t') {
+			prm(p);
+			return 0;
+		} else if (!piseol(p))
+			for (x=0; x!=n; ++x)
+				if (pgetc(p)!=c) {
+					prm(p);
+					return 0;
+				}
+		pnextl(p);
+	}
+	prm(p);
+	return 1;
+}
+
+/* Left indent check */
+/* Verify that there is enough whitespace to do the left indent */
+
+int lindent_check(int c, int n)
+{
+	P *p = pdup(markb);
+	int indwid;
+	if (c=='\t')
+		indwid = n * p->b->o.tab;
+	else
+		indwid = n;
+	while (p->byte < markk->byte) {
+		int x;
+		p_goto_bol(p);
+		if (!piseol(p) && pisindent(p)<indwid) {
+			prm(p);
+			return 0;
+		}
+		pnextl(p);
+	}
+	prm(p);
+	return 1;
+}
+
 /* Indent more */
 
 int urindent(BW *bw)
@@ -585,14 +719,38 @@ int urindent(BW *bw)
 
 			do {
 				pcol(p, markb->xcol);
-				pfill(p, markb->xcol + bw->o.istep, bw->o.indentc == '\t' ? 1 : 0);
+				pfill(p, markb->xcol + bw->o.istep, bw->o.indentc);
 			} while (pnextl(p) && p->line <= markk->line);
 			prm(p);
 		}
 	} else {
-		if (!markb || !markk || markb->b != markk->b || bw->cursor->byte < markb->byte || bw->cursor->byte > markk->byte || markb->byte == markk->byte)
+		if (!markb || !markk || markb->b != markk->b || bw->cursor->byte < markb->byte || bw->cursor->byte > markk->byte || markb->byte == markk->byte) {
 			setindent(bw);
-		else {
+		} else if ( 1 /* bw->o.purify */) {
+			P *p = pdup(markb);
+			P *q = pdup(markb);
+			int indwid;
+
+			if (bw->o.indentc=='\t')
+				indwid = bw->o.tab * bw->o.istep;
+			else
+				indwid = bw->o.istep;
+
+			while (p->byte < markk->byte) {
+				p_goto_bol(p);
+				if (!piseol(p)) {
+					int col;
+					pset(q, p);
+					p_goto_indent(q, bw->o.indentc);
+					col = piscol(q);
+					bdel(p,q);
+					pfill(p,col+indwid,bw->o.indentc);
+				}
+				pnextl(p);
+			}
+			prm(p);
+			prm(q);
+		} else if (purity_check(bw->o.indentc,0)) {
 			P *p = pdup(markb);
 
 			while (p->byte < markk->byte) {
@@ -605,6 +763,10 @@ int urindent(BW *bw)
 				pnextl(p);
 			}
 			prm(p);
+		} else {
+			/* Purity failure */
+			msgnw(bw->parent,US "Selected lines not properly indented");
+			return 1;
 		}
 	}
 	return 0;
@@ -642,27 +804,36 @@ int ulindent(BW *bw)
 			prm(q);
 		}
 	} else {
-		if (!markb || !markk || markb->b != markk->b || bw->cursor->byte < markb->byte || bw->cursor->byte > markk->byte || markb->byte == markk->byte)
+		if (!markb || !markk || markb->b != markk->b || bw->cursor->byte < markb->byte || bw->cursor->byte > markk->byte || markb->byte == markk->byte) {
 			setindent(bw);
-		else {
+		} else if (1 /* bw->o.purify */ && lindent_check(bw->o.indentc,bw->o.istep)) {
+			P *p = pdup(markb);
+			P *q = pdup(markb);
+			int indwid;
+
+			if (bw->o.indentc=='\t')
+				indwid = bw->o.tab * bw->o.istep;
+			else
+				indwid = bw->o.istep;
+
+			while (p->byte < markk->byte) {
+				p_goto_bol(p);
+				if (!piseol(p)) {
+					int col;
+					pset(q, p);
+					p_goto_indent(q, bw->o.indentc);
+					col = piscol(q);
+					bdel(p,q);
+					pfill(p,col-indwid,bw->o.indentc);
+				}
+				pnextl(p);
+			}
+			prm(p);
+			prm(q);
+		} else if (purity_check(bw->o.indentc,bw->o.istep)) {
 			P *p = pdup(markb);
 			P *q = pdup(p);
 
-			p_goto_bol(p);
-			while (p->byte < markk->byte) {
-				if (!piseol(p))
-					while (piscol(p) < bw->o.istep) {
-						int c = pgetc(p);
-
-						if (c != ' ' && c != '\t' && c != bw->o.indentc) {
-							prm(p);
-							prm(q);
-							return -1;
-						}
-					}
-				pnextl(p);
-			}
-			pset(p, markb);
 			p_goto_bol(p);
 			while (p->byte < markk->byte) {
 				if (!piseol(p)) {
@@ -675,6 +846,10 @@ int ulindent(BW *bw)
 			}
 			prm(p);
 			prm(q);
+		} else {
+			/* Purity failure */
+			msgnw(bw->parent,US "Selected lines not properly indented");
+			return 1;
 		}
 	}
 	return 0;
@@ -682,7 +857,7 @@ int ulindent(BW *bw)
 
 /* Insert a file */
 
-int doinsf(BW *bw, char *s, void *object, int *notify)
+int doinsf(BW *bw, unsigned char *s, void *object, int *notify)
 {
 	if (notify)
 		*notify = 1;
@@ -697,7 +872,7 @@ int doinsf(BW *bw, char *s, void *object, int *notify)
 
 			tmp = bload(s);
 			if (error) {
-				msgnw(bw->parent, msgs[error + 5]);
+				msgnw(bw->parent, msgs[-error]);
 				brm(tmp);
 				return -1;
 			}
@@ -721,14 +896,14 @@ int doinsf(BW *bw, char *s, void *object, int *notify)
 			updall();
 			return 0;
 		} else {
-			msgnw(bw->parent, "No block");
+			msgnw(bw->parent, US "No block");
 			return -1;
 	} else {
 		int ret = 0;
 		B *tmp = bload(s);
 
 		if (error) {
-			msgnw(bw->parent, msgs[error + 5]), brm(tmp);
+			msgnw(bw->parent, msgs[-error]), brm(tmp);
 			ret = -1;
 		} else
 			binsb(bw->cursor, tmp);
@@ -743,17 +918,19 @@ int doinsf(BW *bw, char *s, void *object, int *notify)
 
 static int filtflg = 0;
 
-static int dofilt(BW *bw, char *s, void *object, int *notify)
+static int dofilt(BW *bw, unsigned char *s, void *object, int *notify)
 {
 	int fr[2];
 	int fw[2];
+	int flg = 0;
 
 	if (notify)
 		*notify = 1;
-	if (markb && markk && !square && markb->b == bw->b && markk->b == bw->b && markb->byte == markk->byte)
+	if (markb && markk && !square && markb->b == bw->b && markk->b == bw->b && markb->byte == markk->byte) {
+		flg = 1;
 		goto ok;
-	if (!markv(1)) {
-		msgnw(bw->parent, "No block");
+	} if (!markv(1)) {
+		msgnw(bw->parent, US "No block");
 		return -1;
 	}
       ok:
@@ -763,6 +940,10 @@ static int dofilt(BW *bw, char *s, void *object, int *notify)
 	npartial(bw->parent->t->t);
 	ttclsn();
 	if (!fork()) {
+#ifdef HAVE_PUTENV
+		unsigned char		*fname, *name;
+		unsigned	len;
+#endif
 		signrm();
 		close(0);
 		close(1);
@@ -774,14 +955,21 @@ static int dofilt(BW *bw, char *s, void *object, int *notify)
 		close(fr[1]);
 		close(fw[1]);
 		close(fr[0]);
+#ifdef HAVE_PUTENV
+		fname = vsncpy(NULL, 0, sc("JOE_FILENAME="));
+		name = bw->b->name ? bw->b->name : (unsigned char *)"Unnamed";
+		if((len = slen(name)) >= 512)	/* limit filename length */
+			len = 512;
+		fname = vsncpy(sv(fname), name, len);
+		putenv((char *)fname);
+		vsrm(fname);
+#endif
 		execl("/bin/sh", "/bin/sh", "-c", s, NULL);
 		_exit(0);
 	}
 	close(fr[1]);
 	close(fw[0]);
 	if (fork()) {
-		long szz;
-
 		close(fw[1]);
 		if (square) {
 			B *tmp;
@@ -814,10 +1002,17 @@ static int dofilt(BW *bw, char *s, void *object, int *notify)
 			brm(tmp);
 			updall();
 		} else {
-			bdel(markb, markk);
-			szz = markk->b->eof->byte;
-			binsb(markk, bread(fr[0], MAXLONG));
-			pfwrd(markk, markk->b->eof->byte - szz);
+			P *p = pdup(markk);
+			if (!flg)
+				prgetc(p);
+			bdel(markb, p);
+			binsb(p, bread(fr[0], MAXLONG));
+			if (!flg) {
+				pset(p,markk);
+				prgetc(p);
+				bdel(p,markk);
+			}
+			prm(p);
 			if (lightoff)
 				unmark(bw);
 		}
@@ -878,19 +1073,78 @@ int ufilt(BW *bw)
 #else
 	switch (checkmark(bw)) {
 	case 0:
-		if (wmkpw(bw->parent, "Command to filter block through (^C to abort): ", &filthist, dofilt, NULL, NULL, utypebw, NULL, NULL))
+		if (wmkpw(bw->parent, US "Command to filter block through (^C to abort): ", &filthist, dofilt, NULL, NULL, utypebw, NULL, NULL, locale_map))
 			return 0;
 		else
 			return -1;
 	case 1:
-		if (wmkpw(bw->parent, "Command to filter file through (^C to abort): ", &filthist, dofilt, NULL, NULL, utypebw, NULL, NULL))
+		if (wmkpw(bw->parent, US "Command to filter file through (^C to abort): ", &filthist, dofilt, NULL, NULL, utypebw, NULL, NULL, locale_map))
 			return 0;
 		else
 			return -1;
 	case 2:
 	default:
-		msgnw(bw->parent, "No block");
+		msgnw(bw->parent, US "No block");
 		return -1;
 	}
 #endif
 }
+
+/* Force region to lower case */
+
+int ulower(BW *bw)
+{
+	if (markv(1)) {
+		P *q;
+	        P *p;
+	        int c;
+		B *b = bcpy(markb,markk);
+		/* Leave one character in buffer to keep pointers set properly... */
+		q = pdup(markk);
+		prgetc(q);
+		bdel(markb,q);
+		b->o.charmap = markb->b->o.charmap;
+		p=pdup(b->bof);
+		while ((c=pgetc(p))!=NO_MORE_DATA) {
+			c = joe_tolower(b->o.charmap,c);
+			binsc(q,c);
+			pgetc(q);
+		}
+		prm(p);
+		bdel(q,markk);
+		prm(q);
+		brm(b);
+		bw->cursor->xcol = piscol(bw->cursor);
+		return 0;
+	} else
+		return -1;
+}
+
+/* Force region to upper case */
+
+int uupper(BW *bw)
+{
+	if (markv(1)) {
+		P *q;
+	        P *p;
+	        int c;
+		B *b = bcpy(markb,markk);
+		q = pdup(markk);
+		prgetc(q);
+		bdel(markb,q);
+		b->o.charmap = markb->b->o.charmap;
+		p=pdup(b->bof);
+		while ((c=pgetc(p))!=NO_MORE_DATA) {
+			c = joe_toupper(b->o.charmap,c);
+			binsc(q,c);
+			pgetc(q);
+		}
+		prm(p);
+		bdel(q,markk);
+		prm(q);
+		brm(b);
+		bw->cursor->xcol = piscol(bw->cursor);
+		return 0;
+	} else
+		return -1;
+}
