@@ -32,16 +32,21 @@ int ublkdel PARAMS((BW *bw));
 int upicokill PARAMS((BW *bw));
 int ublkmove PARAMS((BW *bw));
 int ublkcpy PARAMS((BW *bw));
-int dowrite PARAMS((BW *bw, char *s, void *object, int *notify));
-int doinsf PARAMS((BW *bw, char *s, void *object, int *notify));
+int dowrite PARAMS((BW *bw, unsigned char *s, void *object, int *notify));
+int doinsf PARAMS((BW *bw, unsigned char *s, void *object, int *notify));
 void setindent PARAMS((BW *bw));
 int urindent PARAMS((BW *bw));
 int ulindent PARAMS((BW *bw));
 int ufilt PARAMS((BW *bw));
 int unmark PARAMS((BW *bw));
 int udrop PARAMS((BW *bw));
+int utoggle_marking PARAMS((BW *bw));
+int ubegin_marking PARAMS((BW *bw));
+int uselect PARAMS((BW *bw));
 int upsh PARAMS((BW *bw));
 int upop PARAMS((BW *bw));
+int ulower PARAMS((BW *bw));
+int uupper PARAMS((BW *bw));
 extern int nstack;
 
 #endif
