@@ -9,7 +9,6 @@
 #include "types.h"
 
 #include <stdio.h>
-#include <ctype.h>
 
 #include "b.h"
 #include "bw.h"
@@ -23,6 +22,8 @@
 #include "umath.h"
 #include "utils.h"
 #include "vs.h"
+#include "utf8.h"
+#include "charmap.h"
 #include "w.h"
 
 /***************/
@@ -40,6 +41,34 @@ int u_goto_bol(BW *bw)
 	return 0;
 }
 
+/*
+ * Move cursor to first non-whitespace character, unless it is
+ * already there, in which case move it to beginning of line
+ */
+int uhome(BW *bw)
+{
+	P *p = pdup(bw->cursor);
+
+	if (bw->o.indentfirst) {
+		if ((bw->o.smarthome) && (piscol(p) > pisindent(p))) { 
+			p_goto_bol(p);
+			while (joe_isblank(p->b->o.charmap,brc(p)))
+				pgetc(p);
+		} else
+			p_goto_bol(p);
+	} else {
+		if (bw->o.smarthome && piscol(p)==0 && pisindent(p)) {
+			while (joe_isblank(p->b->o.charmap,brc(p)))
+				pgetc(p);
+		} else
+			p_goto_bol(p);
+	}
+
+	pset(bw->cursor, p);
+	prm(p);
+	return 0;
+}
+
 /*
  * Move cursor to end of line
  */
@@ -72,10 +101,24 @@ int u_goto_eof(BW *bw)
  */
 int u_goto_left(BW *bw)
 {
-	if (prgetc(bw->cursor) != NO_MORE_DATA)
-		return 0;
-	else
-		return -1;
+	if (bw->o.picture) {
+		if (bw->cursor->xcol) {
+			--bw->cursor->xcol;
+			pcol(bw->cursor,bw->cursor->xcol);
+			return 0;
+		}
+	} else {
+		/* Have to do ECHKXCOL here because of picture mode */
+		if (bw->cursor->xcol != piscol(bw->cursor)) {
+			bw->cursor->xcol = piscol(bw->cursor);
+			return 0;
+		} else if (prgetc(bw->cursor) != NO_MORE_DATA) {
+			bw->cursor->xcol = piscol(bw->cursor);
+			return 0;
+		} else {
+			return -1;
+		}
+	}
 }
 
 /*
@@ -83,10 +126,23 @@ int u_goto_left(BW *bw)
  */
 int u_goto_right(BW *bw)
 {
-	if (pgetc(bw->cursor) != NO_MORE_DATA)
+	if (bw->o.picture) {
+		++bw->cursor->xcol;
+		pcol(bw->cursor,bw->cursor->xcol);
 		return 0;
-	else
-		return -1;
+	} else {
+		int rtn;
+		if (pgetc(bw->cursor) != NO_MORE_DATA) {
+			bw->cursor->xcol = piscol(bw->cursor);
+			rtn = 0;
+		} else {
+			rtn = -1;
+		}
+		/* Have to do EFIXXCOL here because of picture mode */
+		if (bw->cursor->xcol != piscol(bw->cursor))
+			bw->cursor->xcol = piscol(bw->cursor);
+		return rtn;
+	}
 }
 
 /*
@@ -97,23 +153,31 @@ int u_goto_right(BW *bw)
  */
 int u_goto_prev(BW *bw)
 {
-	if (pisbof(bw->cursor)) {
-		return -1;	/* cursor is at beginning of file */
-	} else if (isspace(prgetc(bw->cursor))) {
-		while ((!pisbof(bw->cursor)) && (isspace(prgetc(bw->cursor))))
-			/* do nothing */;	/* if cursor is on white-space char then find first non-white-space char */
-	}
-	if (pisbof(bw->cursor)) {
-		return -1;	/* cursor is at beginning of file */
-	}
-	pgetc(bw->cursor);
+	P *p = pdup(bw->cursor);
+	struct charmap *map=bw->b->o.charmap;
+	int c = prgetc(p);
 
-	while (!pisbof(bw->cursor)) {
-		if (isspace(prgetc(bw->cursor))) {	/* if previous character is white-space then beginning of word was found */
-			pgetc(bw->cursor);
-			break;
-		}
+	if (joe_isalnum_(map,c)) {
+		while (joe_isalnum_(map,(c=prgetc(p))))
+			/* Do nothing */;
+		if (c != NO_MORE_DATA)
+			pgetc(p);
+	} else if (joe_isspace(map,c) || joe_ispunct(map,c)) {
+		while ((c=prgetc(p)), (joe_isspace(map,c) || joe_ispunct(map,c)))
+			/* Do nothing */;
+		while(joe_isalnum_(map,(c=prgetc(p))))
+			/* Do nothing */;
+		if (c != NO_MORE_DATA)
+			pgetc(p);
 	}
+/*
+	if (p->byte == bw->cursor->byte) {
+		prm(p);
+		return -1;
+	}
+*/
+	pset(bw->cursor, p);
+	prm(p);
 	return 0;
 }
 
@@ -125,30 +189,33 @@ int u_goto_prev(BW *bw)
  */
 int u_goto_next(BW *bw)
 {
-	if (piseof(bw->cursor)) {
-		return -1;	/* cursor is at end of file */
-	} else if (isspace(pgetc(bw->cursor))) {
-		while ((!piseof(bw->cursor)) && (isspace(pgetc(bw->cursor))))
-			/* do nothing */;	/* if cursor is on white-space char then find first non-white-space char */
-	}
-	if (piseof(bw->cursor)) {
-		return -1;	/* cursor is at end of file */
-	}
-	prgetc(bw->cursor);
+	P *p = pdup(bw->cursor);
+	struct charmap *map=bw->b->o.charmap;
+	int c = brch(p);
+	int rtn = -1;
 
-	while (!piseof(bw->cursor)) {
-		if (isspace(pgetc(bw->cursor))) {	/* if next character is white-space then end of word was found */
-			prgetc(bw->cursor);
-			break;
+	if (joe_isalnum_(map,c)) {
+		rtn = 0;
+		while (joe_isalnum_(map,(c = brch(p))))
+			pgetc(p);
+	} else if (joe_isspace(map,c) || joe_ispunct(map,c)) {
+		while (joe_isspace(map, (c = brch(p))) || joe_ispunct(map,c))
+			pgetc(p);
+		while (joe_isalnum_(map,(c = brch(p)))) {
+			rtn = 0;
+			pgetc(p);
 		}
-	}
-	return 0;
+	} else
+		pgetc(p);
+	pset(bw->cursor, p);
+	prm(p);
+	return rtn;
 }
 
 static P *pboi(P *p)
 {
 	p_goto_bol(p);
-	while (isblank(brc(p)))
+	while (joe_isblank(p->b->o.charmap,brch(p)))
 		pgetc(p);
 	return p;
 }
@@ -166,9 +233,9 @@ static int pisedge(P *p)
 	pboi(q);
 	if (q->byte == p->byte)
 		goto left;
-	if (isblank(c = brc(p))) {
+	if (joe_isblank(p->b->o.charmap,(c = brch(p)))) {
 		pset(q, p);
-		if (isblank(prgetc(q)))
+		if (joe_isblank(p->b->o.charmap,prgetc(q)))
 			goto no;
 		if (c == '\t')
 			goto right;
@@ -224,7 +291,7 @@ int utomatch(BW *bw)
 	 f,			/* Character to find */
 	 dir;			/* 1 to search forward, -1 to search backward */
 
-	switch (c = brc(bw->cursor)) {
+	switch (c = brch(bw->cursor)) {
 	case '(':
 		f = ')';
 		dir = 1;
@@ -324,6 +391,12 @@ int udnarw(BW *bw)
 		pnextl(bw->cursor);
 		pcol(bw->cursor, bw->cursor->xcol);
 		return 0;
+	} else if(bw->o.picture) {
+		p_goto_eol(bw->cursor);
+		binsc(bw->cursor,'\n');
+		pgetc(bw->cursor);
+		pcol(bw->cursor, bw->cursor->xcol);
+		return 0;
 	} else
 		return -1;
 }
@@ -493,7 +566,7 @@ int udnslide(BW *bw)
 
 static B *linehist = NULL;	/* History of previously entered line numbers */
 
-static int doline(BW *bw, char *s, void *object, int *notify)
+static int doline(BW *bw, unsigned char *s, void *object, int *notify)
 {
 	long num = calc(bw, s);
 
@@ -514,14 +587,14 @@ static int doline(BW *bw, char *s, void *object, int *notify)
 		if (merr)
 			msgnw(bw->parent, merr);
 		else
-			msgnw(bw->parent, "Invalid line number");
+			msgnw(bw->parent, US "Invalid line number");
 		return -1;
 	}
 }
 
 int uline(BW *bw)
 {
-	if (wmkpw(bw->parent, "Go to line (^C to abort): ", &linehist, doline, NULL, NULL, NULL, NULL, NULL))
+	if (wmkpw(bw->parent, US "Go to line (^C to abort): ", &linehist, doline, NULL, NULL, NULL, NULL, NULL, locale_map))
 		return 0;
 	else
 		return -1;
@@ -531,7 +604,7 @@ int uline(BW *bw)
 
 static B *colhist = NULL;	/* History of previously entered column numbers */
 
-static int docol(BW *bw, char *s, void *object, int *notify)
+static int docol(BW *bw, unsigned char *s, void *object, int *notify)
 {
 	long num = calc(bw, s);
 
@@ -550,14 +623,14 @@ static int docol(BW *bw, char *s, void *object, int *notify)
 		if (merr)
 			msgnw(bw->parent, merr);
 		else
-			msgnw(bw->parent, "Invalid column number");
+			msgnw(bw->parent, US "Invalid column number");
 		return -1;
 	}
 }
 
 int ucol(BW *bw)
 {
-	if (wmkpw(bw->parent, "Go to column (^C to abort): ", &colhist, docol, NULL, NULL, NULL, NULL, NULL))
+	if (wmkpw(bw->parent, US "Go to column (^C to abort): ", &colhist, docol, NULL, NULL, NULL, NULL, NULL, locale_map))
 		return 0;
 	else
 		return -1;
@@ -567,7 +640,7 @@ int ucol(BW *bw)
 
 static B *bytehist = NULL;	/* History of previously entered byte numbers */
 
-static int dobyte(BW *bw, char *s, void *object, int *notify)
+static int dobyte(BW *bw, unsigned char *s, void *object, int *notify)
 {
 	long num = calc(bw, s);
 
@@ -586,14 +659,14 @@ static int dobyte(BW *bw, char *s, void *object, int *notify)
 		if (merr)
 			msgnw(bw->parent, merr);
 		else
-			msgnw(bw->parent, "Invalid byte number");
+			msgnw(bw->parent, US "Invalid byte number");
 		return -1;
 	}
 }
 
 int ubyte(BW *bw)
 {
-	if (wmkpw(bw->parent, "Go to byte (^C to abort): ", &bytehist, dobyte, NULL, NULL, NULL, NULL, NULL))
+	if (wmkpw(bw->parent, US "Go to byte (^C to abort): ", &bytehist, dobyte, NULL, NULL, NULL, NULL, NULL, locale_map))
 		return 0;
 	else
 		return -1;
@@ -605,45 +678,88 @@ int ubyte(BW *bw)
 
 int udelch(BW *bw)
 {
-	if (bw->pid && bw->cursor->byte == bw->b->eof->byte) {
-		char c = 4;
+	P *p;
 
-		joe_write(bw->pid, &c, 0);	/* FIXME: why would we write zero chars ??? */
-	} else {
-		P *p;
-
-		if (piseof(bw->cursor))
-			return -1;
-		pgetc(p = pdup(bw->cursor));
-		bdel(bw->cursor, p);
-		prm(p);
-	}
+	if (piseof(bw->cursor))
+		return -1;
+	pgetc(p = pdup(bw->cursor));
+	bdel(bw->cursor, p);
+	prm(p);
 	return 0;
 }
 
-/* Backspace, or if cursor is at end of file in a shell window, send backspace
- * to shell */
+/* Backspace */
 
 int ubacks(BW *bw, int k)
 {
-	if (bw->pid && bw->cursor->byte == bw->b->eof->byte) {
-		char c = k;
-
-		joe_write(bw->out, &c, 1);
-	} else if (bw->parent->watom->what == TYPETW || !pisbol(bw->cursor)) {
+	/* Don't backspace when at beginning of line in prompt windows */
+	if (bw->parent->watom->what == TYPETW || !pisbol(bw->cursor)) {
 		P *p;
 		int c;
+		int indent;
+		int col;
+		int indwid;
+		int wid;
+		int pure = 1;
 
 		if (pisbof(bw->cursor))
 			return -1;
-		p = pdup(bw->cursor);
-		if ((c = prgetc(bw->cursor)) != NO_MORE_DATA)
-			if (!bw->o.overtype || c == '\t' || pisbol(p) || piseol(p))
+
+		/* Indentation point of this line */
+		indent = pisindent(bw->cursor);
+
+		/* Column position of cursor */
+		col = piscol(bw->cursor);
+
+		/* Indentation step in columns */
+		if (bw->o.indentc=='\t')
+			wid = bw->o.tab;
+		else
+			wid = 1;
+
+		indwid = (bw->o.istep*wid);
+
+		/* Smart backspace when: cursor is at indentation point, indentation point
+		   is a multiple of indentation width, we're not at beginning of line,
+		   'smarthome' option is enabled, and indentation is purely made out of
+		   indent characters (or purify indents is enabled). */
+		
+		/* Ignore purify for backspace */
+		if (col == indent && (col%indwid)==0 && col!=0 && bw->o.smartbacks && bw->o.autoindent) {
+			P *p;
+			int x;
+
+			/* Delete all indentation */
+			p = pdup(bw->cursor);
+			p_goto_bol(p);
+			bdel(p,bw->cursor);
+			prm(p);
+
+			/* Indent to new position */
+			pfill(bw->cursor,col-indwid,bw->o.indentc);
+		} else if (col<indent && !pisbol(bw->cursor)) {
+			/* We're before indent point: delete indwid worth of space but do not
+			   cross line boundary.  We could probably replace the above with this. */
+			int cw=0;
+			P *p = pdup(bw->cursor);
+			do {
+				c = prgetc(bw->cursor);
+				if(c=='\t') cw += bw->o.tab;
+				else cw += 1;
 				bdel(bw->cursor, p);
-		prm(p);
+			} while(!pisbol(bw->cursor) && cw<indwid);
+			prm(p);
+		} else {
+			/* Regular backspace */
+			P *p = pdup(bw->cursor);
+			if ((c = prgetc(bw->cursor)) != NO_MORE_DATA)
+				if (!bw->o.overtype || c == '\t' || pisbol(p) || piseol(p))
+					bdel(bw->cursor, p);
+			prm(p);
+		}
+		return 0;
 	} else
 		return -1;
-	return 0;
 }
 
 /* 
@@ -656,13 +772,14 @@ int ubacks(BW *bw, int k)
 int u_word_delete(BW *bw)
 {
 	P *p = pdup(bw->cursor);
-	int c = brc(p);
+	struct charmap *map=bw->b->o.charmap;
+	int c = brch(p);
 
-	if (isalnum_(c))
-		while (isalnum_(c = brc(p)))
+	if (joe_isalnum_(map,c))
+		while (joe_isalnum_(map,(c = brch(p))))
 			pgetc(p);
-	else if (isspace(c))
-		while (isspace(c = brc(p)))
+	else if (joe_isspace(map,c))
+		while (joe_isspace(map,(c = brch(p))))
 			pgetc(p);
 	else
 		pgetc(p);
@@ -684,14 +801,15 @@ int ubackw(BW *bw)
 {
 	P *p = pdup(bw->cursor);
 	int c = prgetc(bw->cursor);
+	struct charmap *map=bw->b->o.charmap;
 
-	if (isalnum_(c)) {
-		while (isalnum_(c = prgetc(bw->cursor)))
+	if (joe_isalnum_(map,c)) {
+		while (joe_isalnum_(map,(c = prgetc(bw->cursor))))
 			/* do nothing */;
 		if (c != NO_MORE_DATA)
 			pgetc(bw->cursor);
-	} else if (isspace(c)) {
-		while (isspace(c = prgetc(bw->cursor)))
+	} else if (joe_isspace(map,c)) {
+		while (joe_isspace(map,(c = prgetc(bw->cursor))))
 			/* do nothing */;
 		if (c != NO_MORE_DATA)
 			pgetc(bw->cursor);
@@ -732,7 +850,7 @@ int udelbl(BW *bw)
 
 	if (p->byte == bw->cursor->byte) {
 		prm(p);
-		return ubacks(bw, MAXINT);	/* FIXME: MAXINT overloaded */
+		return ubacks(bw, 8);	/* The 8 goes to the process if we're at EOF of shell window */
 	} else
 		bdel(p, bw->cursor);
 	prm(p);
@@ -764,55 +882,148 @@ int uinsc(BW *bw)
 	return 0;
 }
 
+/* Move p backwards to first non-blank line and return its indentation */
+
+int find_indent(P *p)
+{
+	int x;
+	for (x=0; x != 10; ++x) {
+		if (!pprevl(p)) return -1;
+		p_goto_bol(p);
+		if (!pisblank(p)) break;
+	}
+	if (x==10)
+		return -1;
+	else
+		return pisindent(p);
+}
+
 /* Type a character into the buffer (deal with left margin, overtype mode and
  * word-wrap), if cursor is at end of shell window buffer, just send character
  * to process.
  */
 
+struct utf8_sm utype_utf8_sm;
+
 int utypebw(BW *bw, int k)
 {
-	if (bw->pid && bw->cursor->byte == bw->b->eof->byte) {
-		char c = k;
-
-		joe_write(bw->out, &c, 1);
+	struct charmap *map=bw->b->o.charmap;
+	if (k == '\t' && bw->o.overtype && !piseol(bw->cursor)) { /* TAB in overtype mode is supposed to be just cursor motion */
+		int col = bw->cursor->xcol;		/* Current cursor column */
+		col = col + bw->o.tab - (col%bw->o.tab);/* Move to next tab stop */
+		pcol(bw->cursor,col);			/* Try to position cursor there */
+		if (!bw->o.picture && piseol(bw->cursor) && piscol(bw->cursor)<col)	/* We moved past end of line, insert a tab (unless in picture mode) */
+			if (bw->o.spaces)
+				pfill(bw->cursor,col,' ');
+			else
+				pfill(bw->cursor,col,'\t');
+		bw->cursor->xcol = col;			/* Put cursor there even if we can't really go there */
+	} else if (k == '\t' && bw->o.smartbacks && bw->o.autoindent && pisindent(bw->cursor)>=piscol(bw->cursor)) {
+		P *p = pdup(bw->cursor);
+		int n = find_indent(p);
+		if (n != -1 && pisindent(bw->cursor)==piscol(bw->cursor) && n > pisindent(bw->cursor)) {
+			if (!pisbol(bw->cursor))
+				udelbl(bw);
+			while (joe_isspace(map,(k = pgetc(p))) && k != '\n') {
+				binsc(bw->cursor, k);
+				pgetc(bw->cursor);
+			}
+		} else {
+			int x;
+			for (x=0;x<bw->o.istep;++x) {
+				binsc(bw->cursor,bw->o.indentc);
+				pgetc(bw->cursor);
+			}
+		}
+		bw->cursor->xcol = piscol(bw->cursor);
+		prm (p);
 	} else if (k == '\t' && bw->o.spaces) {
-		long n = piscol(bw->cursor);
+		long n;
+
+		if (bw->o.picture)
+			n = bw->cursor->xcol;
+		else
+			n = piscol(bw->cursor);
+
+		utype_utf8_sm.state = 0;
+		utype_utf8_sm.ptr = 0;
 
 		n = bw->o.tab - n % bw->o.tab;
 		while (n--)
 			utypebw(bw, ' ');
 	} else {
-		int upd = bw->parent->t->t->updtab[bw->y + bw->cursor->line - bw->top->line];
-		int simple = 1;
+		int upd;
+		int simple;
+		int x;
+
+		/* Picture mode */
+		if (bw->o.picture && bw->cursor->xcol!=piscol(bw->cursor))
+			pfill(bw->cursor,bw->cursor->xcol,' '); /* Why no tabs? */
+
+		/* UTF8 decoder */
+		if(locale_map->type) {
+			int utf8_char = utf8_decode(&utype_utf8_sm,k);
+
+			if(utf8_char >= 0)
+				k = utf8_char;
+			else
+				return 0;
+		}
+
+		upd = bw->parent->t->t->updtab[bw->y + bw->cursor->line - bw->top->line];
+		simple = 1;
 
 		if (pisblank(bw->cursor))
 			while (piscol(bw->cursor) < bw->o.lmargin) {
 				binsc(bw->cursor, ' ');
 				pgetc(bw->cursor);
 			}
-		binsc(bw->cursor, k), pgetc(bw->cursor);
-		if (bw->o.wordwrap && piscol(bw->cursor) > bw->o.rmargin && !isblank(k)) {
+
+		if(locale_map->type && !bw->b->o.charmap->type) {
+			unsigned char buf[10];
+			utf8_encode(buf,k);
+			k = from_utf8(bw->b->o.charmap,buf);
+		} else if(!locale_map->type && bw->b->o.charmap->type) {
+			unsigned char buf[10];
+			to_utf8(locale_map,buf,k);
+			k = utf8_decode_string(buf);
+		}
+		
+		binsc(bw->cursor, k);
+
+		/* We need x position before we move cursor */
+		x = piscol(bw->cursor) - bw->offset;
+		pgetc(bw->cursor);
+
+		/* Tabs are weird here... */
+		if (bw->o.overtype && !piseol(bw->cursor) && k != '\t')
+			udelch(bw);
+
+		/* Not sure if we're in right position for wordwrap when we're in overtype mode */
+		if (bw->o.wordwrap && piscol(bw->cursor) > bw->o.rmargin && !joe_isblank(map,k)) {
 			wrapword(bw->cursor, (long) bw->o.lmargin, bw->o.french, NULL);
 			simple = 0;
-		} else if (bw->o.overtype && !piseol(bw->cursor) && k != '\t')
-			udelch(bw);
+		}
+
 		bw->cursor->xcol = piscol(bw->cursor);
 #ifndef __MSDOS__
-		if (bw->cursor->xcol - bw->offset - 1 < 0 || bw->cursor->xcol - bw->offset - 1 >= bw->w)
+		if (x < 0 || x >= bw->w)
 			simple = 0;
 		if (bw->cursor->line < bw->top->line || bw->cursor->line >= bw->top->line + bw->h)
 			simple = 0;
 		if (simple && bw->parent->t->t->sary[bw->y + bw->cursor->line - bw->top->line])
 			simple = 0;
 		if (simple && k != '\t' && k != '\n' && !curmacro) {
-			int a = 0;
-			unsigned char c = (unsigned char)k;
+			int a;
+			int atr = 0;
+			unsigned char c = k;
 			SCRN *t = bw->parent->t->t;
 			int y = bw->y + bw->cursor->line - bw->top->line;
-			int x = bw->cursor->xcol - bw->offset + bw->x - 1;
 			int *screen = t->scrn + y * t->co;
+			int *attr = t->attr + y * t->co;
+			x += bw->x;
 
-			if (!upd && piseol(bw->cursor))
+			if (!upd && piseol(bw->cursor) && !bw->o.highlight)
 				t->updtab[y] = 0;
 			if (markb &&
 			    markk &&
@@ -820,9 +1031,8 @@ int utypebw(BW *bw, int k)
 			    markk->b == bw->b &&
 			   ((!square && bw->cursor->byte >= markb->byte && bw->cursor->byte < markk->byte) ||
 			    ( square && bw->cursor->line >= markb->line && bw->cursor->line <= markk->line && piscol(bw->cursor) >= markb->xcol && piscol(bw->cursor) < markk->xcol)))
-				a = INVERSE;
-			xlat(&a, &c);
-			outatr(t, screen + x, x, y, c, a);
+				atr = INVERSE;
+			outatr(bw->b->o.charmap, t, screen + x, attr + x, x, y, k, atr);
 		}
 #endif
 	}
@@ -831,12 +1041,30 @@ int utypebw(BW *bw, int k)
 
 /* Quoting */
 
+static B *unicodehist = NULL;	/* History of previously entered unicode characters */
+
+static int dounicode(BW *bw, unsigned char *s, void *object, int *notify)
+{
+	int num;
+	unsigned char buf[8];
+	int x;
+	sscanf((char *)s,"%x",&num);
+	if (notify)
+		*notify = 1;
+	vsrm(s);
+	utf8_encode(buf,num);
+	for(x=0;buf[x];++x)
+		utypebw(bw, buf[x]);
+	bw->cursor->xcol = piscol(bw->cursor);
+	return 0;
+}
+
 int quotestate;
 int quoteval;
 
 static int doquote(BW *bw, int c, void *object, int *notify)
 {
-	char buf[40];
+	unsigned char buf[40];
 
 	if (c < 0 || c >= 256) {
 		nungetc(c);
@@ -847,17 +1075,25 @@ static int doquote(BW *bw, int c, void *object, int *notify)
 		if (c >= '0' && c <= '9') {
 			quoteval = c - '0';
 			quotestate = 1;
-			snprintf(buf, sizeof(buf), "ASCII %c--", c);
+			joe_snprintf_1((char *)buf, sizeof(buf), "ASCII %c--", c);
 			if (!mkqwna(bw->parent, sz(buf), doquote, NULL, NULL, notify))
 				return -1;
 			else
 				return 0;
 		} else if (c == 'x' || c == 'X') {
-			quotestate = 3;
-			if (!mkqwna(bw->parent, sc("ASCII 0x--"), doquote, NULL, NULL, notify))
-				return -1;
-			else
-				return 0;
+			if (bw->b->o.charmap->type) {
+				if (!wmkpw(bw->parent, US "Unicode (ISO-10646) character in hex (^C to abort): ", &unicodehist, dounicode,
+				           NULL, NULL, NULL, NULL, NULL, locale_map))
+					return 0;
+				else
+					return -1;
+			} else {
+				quotestate = 3;
+				if (!mkqwna(bw->parent, sc("ASCII 0x--"), doquote, NULL, NULL, notify))
+					return -1;
+				else
+					return 0;
+			}
 		} else if (c == 'o' || c == 'O') {
 			quotestate = 5;
 			if (!mkqwna(bw->parent, sc("ASCII 0---"), doquote, NULL, NULL, notify))
@@ -875,7 +1111,7 @@ static int doquote(BW *bw, int c, void *object, int *notify)
 		break;
 	case 1:
 		if (c >= '0' && c <= '9') {
-			snprintf(buf, sizeof(buf), "ASCII %c%c-", quoteval + '0', c);
+			joe_snprintf_2((char *)buf, sizeof(buf), "ASCII %c%c-", quoteval + '0', c);
 			quoteval = quoteval * 10 + c - '0';
 			quotestate = 2;
 			if (!mkqwna(bw->parent, sz(buf), doquote, NULL, NULL, notify))
@@ -893,7 +1129,7 @@ static int doquote(BW *bw, int c, void *object, int *notify)
 		break;
 	case 3:
 		if (c >= '0' && c <= '9') {
-			snprintf(buf, sizeof(buf), "ASCII 0x%c-", c);
+			joe_snprintf_1((char *)buf, sizeof(buf), "ASCII 0x%c-", c);
 			quoteval = c - '0';
 			quotestate = 4;
 			if (!mkqwna(bw->parent, sz(buf), doquote, NULL, NULL, notify))
@@ -901,7 +1137,7 @@ static int doquote(BW *bw, int c, void *object, int *notify)
 			else
 				return 0;
 		} else if (c >= 'a' && c <= 'f') {
-			snprintf(buf, sizeof(buf), "ASCII 0x%c-", c + 'A' - 'a');
+			joe_snprintf_1((char *)buf, sizeof(buf), "ASCII 0x%c-", c + 'A' - 'a');
 			quoteval = c - 'a' + 10;
 			quotestate = 4;
 			if (!mkqwna(bw->parent, sz(buf), doquote, NULL, NULL, notify))
@@ -909,7 +1145,7 @@ static int doquote(BW *bw, int c, void *object, int *notify)
 			else
 				return 0;
 		} else if (c >= 'A' && c <= 'F') {
-			snprintf(buf, sizeof(buf), "ASCII 0x%c-", c);
+			joe_snprintf_1((char *)buf, sizeof(buf), "ASCII 0x%c-", c);
 			quoteval = c - 'A' + 10;
 			quotestate = 4;
 			if (!mkqwna(bw->parent, sz(buf), doquote, NULL, NULL, notify))
@@ -935,7 +1171,7 @@ static int doquote(BW *bw, int c, void *object, int *notify)
 		break;
 	case 5:
 		if (c >= '0' && c <= '7') {
-			snprintf(buf, sizeof(buf), "ASCII 0%c--", c);
+			joe_snprintf_1((char *)buf, sizeof(buf), "ASCII 0%c--", c);
 			quoteval = c - '0';
 			quotestate = 6;
 			if (!mkqwna(bw->parent, sz(buf), doquote, NULL, NULL, notify))
@@ -946,7 +1182,7 @@ static int doquote(BW *bw, int c, void *object, int *notify)
 		break;
 	case 6:
 		if (c >= '0' && c <= '7') {
-			snprintf(buf, sizeof(buf), "ASCII 0%c%c-", quoteval + '0', c);
+			joe_snprintf_2((char *)buf, sizeof(buf), "ASCII 0%c%c-", quoteval + '0', c);
 			quoteval = quoteval * 8 + c - '0';
 			quotestate = 7;
 			if (!mkqwna(bw->parent, sz(buf), doquote, NULL, NULL, notify))
@@ -1015,7 +1251,7 @@ int uquote8(BW *bw)
 		return -1;
 }
 
-extern char srchstr[];
+extern unsigned char srchstr[];
 
 static int doctrl(BW *bw, int c, void *object, int *notify)
 {
@@ -1042,22 +1278,26 @@ int uctrl(BW *bw)
 		return -1;
 }
 
-/* User hit Return.  Deal with autoindent.  If cursor is at end of shell
- * window buffer, send the return to the shell
+/* User hit Return.  Deal with autoindent.
  */
 
 int rtntw(BW *bw)
 {
-	if (bw->pid && bw->cursor->byte == bw->b->eof->byte) {
-		joe_write(bw->out, "\n", 1);
+	if (bw->o.overtype) {
+		p_goto_eol(bw->cursor);
+		if (piseof(bw->cursor))
+			binsc(bw->cursor, '\n');
+		pgetc(bw->cursor);
+		bw->cursor->xcol = piscol(bw->cursor);
 	} else {
 		P *p = pdup(bw->cursor);
-		char c;
+		unsigned char c;
 
 		binsc(bw->cursor, '\n'), pgetc(bw->cursor);
-		if (bw->o.autoindent) {
+		/* Suppress autoindent if we're on a space or tab... */
+		if (bw->o.autoindent && (brch(bw->cursor)!=' ' && brch(bw->cursor)!='\t')) {
 			p_goto_bol(p);
-			while (isspace(c = pgetc(p)) && c != 10) {
+			while (joe_isspace(bw->b->o.charmap,(c = pgetc(p))) && c != '\n') {
 				binsc(bw->cursor, c);
 				pgetc(bw->cursor);
 			}
@@ -1072,11 +1312,7 @@ int rtntw(BW *bw)
 
 int uopen(BW *bw)
 {
-	P *q = pdup(bw->cursor);
-
-	rtntw(bw);
-	pset(bw->cursor, q);
-	prm(q);
+	binsc(bw->cursor,'\n');
 	return 0;
 }
 
@@ -1086,11 +1322,13 @@ static int dosetmark(BW *bw, int c, void *object, int *notify)
 {
 	if (notify)
 		*notify = 1;
-	if (c >= '0' && c <= '9') {
+	if (c >= '0' && c <= ':') {
 		pdupown(bw->cursor, bw->b->marks + c - '0');
 		poffline(bw->b->marks[c - '0']);
-		snprintf(msgbuf, JOE_MSGBUFSIZE, "Mark %d set", c - '0');
-		msgnw(bw->parent, msgbuf);
+		if (c!=':') {
+			joe_snprintf_1((char *)msgbuf, JOE_MSGBUFSIZE, "Mark %d set", c - '0');
+			msgnw(bw->parent, msgbuf);
+		}
 		return 0;
 	} else {
 		nungetc(c);
@@ -1100,7 +1338,7 @@ static int dosetmark(BW *bw, int c, void *object, int *notify)
 
 int usetmark(BW *bw, int c)
 {
-	if (c >= '0' && c <= '9')
+	if (c >= '0' && c <= ':')
 		return dosetmark(bw, c, NULL, NULL);
 	else if (mkqwna(bw->parent, sc("Set mark (0-9):"), dosetmark, NULL, NULL, NULL))
 		return 0;
@@ -1114,13 +1352,13 @@ static int dogomark(BW *bw, int c, void *object, int *notify)
 {
 	if (notify)
 		*notify = 1;
-	if (c >= '0' && c <= '9')
+	if (c >= '0' && c <= ':')
 		if (bw->b->marks[c - '0']) {
 			pset(bw->cursor, bw->b->marks[c - '0']);
 			bw->cursor->xcol = piscol(bw->cursor);
 			return 0;
 		} else {
-			snprintf(msgbuf, JOE_MSGBUFSIZE, "Mark %d not set", c - '0');
+			joe_snprintf_1((char *)msgbuf, JOE_MSGBUFSIZE, "Mark %d not set", c - '0');
 			msgnw(bw->parent, msgbuf);
 			return -1;
 	} else {
@@ -1165,7 +1403,7 @@ static int dofwrdc(BW *bw, int k, void *object, int *notify)
 				break;
 	}
 	if (c == NO_MORE_DATA) {
-		msgnw(bw->parent, "Not found");
+		msgnw(bw->parent, US "Not found");
 		prm(q);
 		return -1;
 	} else {
@@ -1200,11 +1438,11 @@ int ubkwdc(BW *bw, int k)
 
 /* Display a message */
 
-static int domsg(BASE *b, char *s, void *object, int *notify)
+static int domsg(BASE *b, unsigned char *s, void *object, int *notify)
 {
 	if (notify)
 		*notify = 1;
-	strcpy(msgbuf, s);
+	strcpy((char *)msgbuf, (char *)s);
 	vsrm(s);
 	msgnw(b->parent, msgbuf);
 	return 0;
@@ -1212,7 +1450,7 @@ static int domsg(BASE *b, char *s, void *object, int *notify)
 
 int umsg(BASE *b)
 {
-	if (wmkpw(b->parent, "Msg (^C to abort): ", NULL, domsg, NULL, NULL, NULL, NULL, NULL))
+	if (wmkpw(b->parent, US "Msg (^C to abort): ", NULL, domsg, NULL, NULL, NULL, NULL, NULL, locale_map))
 		return 0;
 	else
 		return -1;
@@ -1220,7 +1458,7 @@ int umsg(BASE *b)
 
 /* Insert text */
 
-static int dotxt(BW *bw, char *s, void *object, int *notify)
+static int dotxt(BW *bw, unsigned char *s, void *object, int *notify)
 {
 	int x;
 
@@ -1232,9 +1470,9 @@ static int dotxt(BW *bw, char *s, void *object, int *notify)
 	return 0;
 }
 
-int utxt(BASE *bw)
+int utxt(BW *bw)
 {
-	if (wmkpw(bw->parent, "Insert (^C to abort): ", NULL, dotxt, NULL, NULL, utypebw, NULL, NULL))
+	if (wmkpw(bw->parent, US "Insert (^C to abort): ", NULL, dotxt, NULL, NULL, utypebw, NULL, NULL, bw->b->o.charmap))
 		return 0;
 	else
 		return -1;
