@@ -18,6 +18,7 @@
 #include "ufile.h"
 #include "utils.h"
 #include "vs.h"
+#include "charmap.h"
 #include "w.h"
 
 /* Error database */
@@ -28,9 +29,9 @@ struct error {
 	LINK(ERROR) link;	/* Linked list of errors */
 	long line;		/* Target line number */
 	long org;		/* Original target line number */
-	char *file;		/* Target file name */
+	unsigned char *file;		/* Target file name */
 	long src;		/* Error-file line number */
-	char *msg;		/* The message */
+	unsigned char *msg;		/* The message */
 } errors = { { &errors, &errors} };
 ERROR *errptr = &errors;	/* Current error row */
 
@@ -38,7 +39,7 @@ B *errbuf = NULL;		/* Buffer with error messages */
 
 /* Insert and delete notices */
 
-void inserr(char *name, long int where, long int n, int bol)
+void inserr(unsigned char *name, long int where, long int n, int bol)
 {
 	ERROR *e;
 
@@ -54,7 +55,7 @@ void inserr(char *name, long int where, long int n, int bol)
 	}
 }
 
-void delerr(char *name, long int where, long int n)
+void delerr(unsigned char *name, long int where, long int n)
 {
 	ERROR *e;
 
@@ -72,7 +73,7 @@ void delerr(char *name, long int where, long int n)
 
 /* Abort notice */
 
-void abrerr(char *name)
+void abrerr(unsigned char *name)
 {
 	ERROR *e;
 
@@ -84,7 +85,7 @@ void abrerr(char *name)
 
 /* Save notice */
 
-void saverr(char *name)
+void saverr(unsigned char *name)
 {
 	ERROR *e;
 
@@ -117,17 +118,29 @@ static void freeall(void)
 
 /* Parse error messages into database */
 
-static int parseit(char *s, long int row)
+/* From joe's joe 2.9 */
+
+/* First word on line with a '.' in it.  This is the file name.  The next number after that is the line number. */
+
+static int parseit(struct charmap *map,unsigned char *s, long int row)
 {
-	int x, y;
-	char *name = NULL;
+	int x, y, flg;
+	unsigned char *name = NULL;
 	long line = -1;
 	ERROR *err;
 
-	/* Skip to first word */
-	for (x = 0; s[x] && !(isalnum_(s[x]) || s[x] == '.' || s[x] == '/'); ++x) ;
-	/* Skip to end of first word */
-	for (y = x; isalnum_(s[y]) || s[y] == '.' || s[y] == '/'; ++y) ;
+	y=0;
+	flg=0;
+
+	do {
+		/* Skip to first word */
+		for (x = y; s[x] && !(joe_isalnum_(map,s[x]) || s[x] == '.' || s[x] == '/'); ++x) ;
+
+		/* Skip to end of first word */
+		for (y = x; joe_isalnum_(map,s[y]) || s[y] == '.' || s[y] == '/'; ++y)
+			if (s[y] == '.')
+				flg = 1;
+	} while (!flg && x!=y);
 
 	/* Save file name */
 	if (x != y)
@@ -141,12 +154,22 @@ static int parseit(char *s, long int row)
 
 	/* Save line number */
 	if (x != y)
-		sscanf(s + x, "%ld", &line);
+		sscanf((char *)(s + x), "%ld", &line);
 	if (line != -1)
 		--line;
 
+	/* Look for ':' */
+	flg = 0;
+	while (s[y]) {
+		if (s[y]==':') {
+			flg = 1;
+			break;
+		}
+		++y;
+	}
+
 	if (name) {
-		if (line != -1) {
+		if (line != -1 && flg) {
 			/* We have an error */
 			err = (ERROR *) alitem(&errnodes, sizeof(ERROR));
 			err->file = name;
@@ -172,13 +195,13 @@ static long parserr(B *b)
 
 	freeall();
 	do {
-		char *s;
+		unsigned char *s;
 
 		pset(q, p);
 		p_goto_eol(p);
 		s = brvs(q, (int) (p->byte - q->byte));
 		if (s) {
-			nerrs += parseit(s, q->line);
+			nerrs += parseit(b->o.charmap, s, q->line);
 			vsrm(s);
 		}
 	} while (pgetc(p) != NO_MORE_DATA);
@@ -187,11 +210,57 @@ static long parserr(B *b)
 	return nerrs;
 }
 
+BW *find_a_good_bw(B *b)
+{
+	W *w;
+	BW *bw = 0;
+	/* Find lowest window with buffer */
+	if ((w = maint->topwin) != NULL) {
+		do {
+			if ((w->watom->what&TYPETW) && ((BW *)w->object)->b==b && w->y>=0)
+				bw = (BW *)w->object;
+			w = w->link.next;
+		} while (w != maint->topwin);
+	}
+	if (bw)
+		return bw;
+	/* Otherwise just find lowest window */
+	if ((w = maint->topwin) != NULL) {
+		do {
+			if ((w->watom->what&TYPETW) && w->y>=0)
+				bw = (BW *)w->object;
+			w = w->link.next;
+		} while (w != maint->topwin);
+	}
+	return bw;
+}
+
+int parserrb(B *b)
+{
+	BW *bw;
+	int n;
+	errbuf = b;
+	freeall();
+	n = parserr(b);
+	bw = find_a_good_bw(b);
+	if (n)
+		joe_snprintf_1((char *)msgbuf, JOE_MSGBUFSIZE, "%ld messages found", n);
+	else
+		joe_snprintf_0((char *)msgbuf, JOE_MSGBUFSIZE, "No messages found");
+	msgnw(bw->parent, msgbuf);
+	return 0;
+}
+
 int uparserr(BW *bw)
 {
+	int n;
 	errbuf = bw->b;
 	freeall();
-	snprintf(msgbuf, JOE_MSGBUFSIZE, "Parsed %ld lines", parserr(bw->b));
+	n = parserr(bw->b);
+	if (n)
+		joe_snprintf_1((char *)msgbuf, JOE_MSGBUFSIZE, "%ld messages found", n);
+	else
+		joe_snprintf_0((char *)msgbuf, JOE_MSGBUFSIZE, "No messages found");
 	msgnw(bw->parent, msgbuf);
 	return 0;
 }
@@ -201,7 +270,7 @@ int unxterr(BW *bw)
 	int omid;
 
 	if (errptr->link.next == &errors) {
-		msgnw(bw->parent, "No more errors");
+		msgnw(bw->parent, US "No more errors");
 		return -1;
 	}
 	errptr = errptr->link.next;
@@ -226,7 +295,7 @@ int uprverr(BW *bw)
 	int omid;
 
 	if (errptr->link.prev == &errors) {
-		msgnw(bw->parent, "No more errors");
+		msgnw(bw->parent, US "No more errors");
 		return -1;
 	}
 	errptr = errptr->link.prev;
