@@ -13,10 +13,11 @@
 
 int unxterr PARAMS((BW *bw));
 int uprverr PARAMS((BW *bw));
+int parserrb PARAMS((B *b));
 int uparserr PARAMS((BW *bw));
-void inserr PARAMS((char *name, long int where, long int n, int bol));
-void delerr PARAMS((char *name, long int where, long int n));
-void abrerr PARAMS((char *name));
-void saverr PARAMS((char *name));
+void inserr PARAMS((unsigned char *name, long int where, long int n, int bol));
+void delerr PARAMS((unsigned char *name, long int where, long int n));
+void abrerr PARAMS((unsigned char *name));
+void saverr PARAMS((unsigned char *name));
 
 #endif
