@@ -13,12 +13,14 @@
 
 extern int exask;
 
-void genexmsg PARAMS((BW *bw, int saved, char *name));
+void genexmsg PARAMS((BW *bw, int saved, unsigned char *name));
 
 int ublksave PARAMS((BW *bw));
 int ushell PARAMS((BW *bw));
 int usave PARAMS((BW *bw));
 int uedit PARAMS((BW *bw));
+int uswitch PARAMS((BW *bw));
+int uscratch PARAMS((BW *bw));
 int uinsf PARAMS((BW *bw));
 int uexsve PARAMS((BW *bw));
 int unbuf PARAMS((BW *bw));
@@ -27,6 +29,8 @@ int uask PARAMS((BW *bw));
 int ubufed PARAMS((BW *bw));
 int ulose PARAMS((BW *bw));
 int okrepl PARAMS((BW *bw));
-int doedit PARAMS((BW *bw, char *s, void *obj, int *notify));
+int doedit PARAMS((BW *bw, unsigned char *s, void *obj, int *notify));
+int uquerysave PARAMS((BW *bw));
+int ukilljoe PARAMS((BW *bw));
 
 #endif
